/**
    @file ft51_general_config.h

    @brief General configuration library

    Functions to initialise, reset, identify the chip.

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

#ifndef __FT51_GENERAL_CONFIG_H__
#define __FT51_GENERAL_CONFIG_H__

#include "ft51.h"
#include "ft51_io_registers.h"
#include "registers/ft51_registers.h"
#include "registers/ft51_mtp_registers.h"

/**
    @enum  FT51_PACKAGE
    @brief FT51 IC packages.
**/
typedef enum
{
    /** IC package = 24 pin SSOP **/
    FT51_24_SSOP,
    /** IC package = 32 pin QFN **/
    FT51_32_QFN,
    /** IC package = 44 pin TQFP **/
    FT51_44_TQFP,
    /** IC package = 48 pin QFN **/
    FT51_48_QFN,
} FT51_PACKAGE;

/**
    @enum  FT51_SYSTEM_CLOCK
    @brief FT51 System clock values.
**/
typedef enum
{
    FT51_6MHZ = 6, ///< System Clock = 6MHz
    FT51_12MHZ = 12, ///< System Clock = 12MHz
    FT51_24MHZ = 24, ///< System Clock = 24MHz
    FT51_48MHZ = 48, ///< System Clock = 48MHz
} FT51_SYSTEM_CLOCK;

/**
    @brief          get_ic_package

    Allows the user to identify which package the firmware is
                    running on.
    @return         Unique identifier for the current I.C. package
**/
FT51_PACKAGE get_ic_package(void);

/**
    @brief          get_system_clock

    Gets the current value of the system clock
                    either 6, 12, 24, or 48 MHz.

    @return         Current frequency of the system clock.
**/
FT51_SYSTEM_CLOCK get_system_clock(void);

/**
    @brief          set_system_clock

    Sets the system clock to either 6, 12, 24, or 48 MHz.

    @param[in]      system_clock    Frequency to set the clock to.
**/
void set_system_clock(const FT51_SYSTEM_CLOCK system_clock);

/**
    @brief          device_stop_request

    Perform a system stop request putting the device
                    into a low power mode.
**/
#define device_stop_request() \
    do { \
        IO_REG_GENERAL_SET_MASK(SYSTEM_CLOCK_DIVIDER, MASK_SYSTEM_STOP_REQUEST); \
    } while (0)

/**
    @brief          device_reset

    Perform a system reset by setting all registers
                    to default values.
                    Does not reload program from MTP.
**/
#define device_reset() \
    do { \
        IO_REG_GENERAL_WRITE(DEVICE_CONTROL_REGISTER, (MASK_TOP_DEV_EN)); \
        IO_REG_GENERAL_WRITE(SYSTEM_CLOCK_DIVIDER, (MASK_RESET_8051)); \
    } while (0)

/**
    @brief          device_revert

    Reloads firmware from MTP and resets all registers
                    to default values.
**/
#define device_revert() \
    do { \
        IO_REG_GENERAL_WRITE(DEVICE_CONTROL_REGISTER, (MASK_TOP_DEV_EN)); \
        IO_REG_GENERAL_WRITE(MTP_CONTROL, (MASK_COPY_MTP_2_RAM)); \
    } while (0)

/**
    @brief          device_sleep

    Puts the device into low power mode. Caller responsible for
                    powering down individual peripherals.
**/
#define device_sleep() \
    do { \
        IO_REG_GENERAL_WRITE(DEVICE_CONTROL_REGISTER, (MASK_TOP_DEV_EN)); \
        IO_REG_GENERAL_WRITE(SYSTEM_CLOCK_DIVIDER, (MASK_SYSTEM_STOP_REQUEST)); \
    } while (0)

/**
    @brief    Initialise FT51

    Enables access to on-chip peripherals such as SPI and UART.
    Also enables debug messages on serial port.
    @return   N/A
**/
void device_initialise(void);

#endif // __FT51_GENERAL_CONFIG_H__
