/**
    @file mouse_app.c

    @brief This will emulate a mouse, allowing the user to control the 
          direction and speed of the mouse pointer.
 
	@details This will emulate a mouse, allowing the user to control the 
	direction and speed of the mouse pointer
 
  - Movement of mouse is controlled by the two switches
       - left  switch only -> move left or up
       - right switch only -> move right or down
       - both switches     -> switch between left/right and up/down movement
       - speed of mouse increases according to duration of switch press (until maximum speed is reached)
 
  - Left mouse button is simulated using the the pressure sensor
       - when force goes above predefined limit the button is *

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */
 
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>

#include "registers\ft51_aio_registers.h"
#include "registers\ft51_iocell_registers.h"

#include "ft51_iomux.h"
#include "ft51.h"
#include "ft51_general_config.h"
#include "ft51_interrupts.h"
#include "ft51_usb.h"
#include "ft51_usb_dfu.h"
#include "ft51_usb_hid.h"
#include "ft51_i2c_master.h"
#include "ft51_lcd.h"
#include "ft51_dfu.h"

// --------------------------------------------------------------------------------------
// Various globals
// --------------------------------------------------------------------------------------
#define MOUSE_REPORT_PERIOD 25 // how often mouse reports should be sent (milliseconds)
#define MOUSE_SPEED_LIMIT 75   // maximum speed the mouse can be move each time
#define MOUSE_SPEED_SCALE 5    // scales the mouse speed down

static volatile int16_t  ms_timer     = 0; // millisecond countdown timer - general purpose
static volatile int16_t  mouse_timer  = 0; // millisecond countdown timer - for sending mouse reports
static volatile uint8_t  mouse_button = 0; // what switches have been pressed (used for mouse movement)
static volatile uint16_t mouse_force  = 0; // current force on sensor (used for mouse button 1)

/**
 @name USB and Hub Configuration
 @brief Indication of whether the USB Hub is enabled, how it is
        powered and the size of the control endpoint Max Packets.
 */
//@{
// USB Hub Enabled - set to 1 for enabled and 0 for disabled
#ifndef USB_HUB_ENABLED
#define USB_HUB_ENABLED 0
#endif // USB_HUB_ENABLED
// USB Bus Powered - set to 1 for self-powered or 0 for bus-powered
#ifndef USB_SELF_POWERED
#define USB_SELF_POWERED 0
#endif // USB_SELF_POWERED
#if USB_SELF_POWERED == 1
#define USB_CONFIG_BMATTRIBUTES_VALUE (USB_CONFIG_BMATTRIBUTES_SELF_POWERED | USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1)
#else // USB_SELF_POWERED
#define USB_CONFIG_BMATTRIBUTES_VALUE USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1
#endif // USB_SELF_POWERED
// USB Endpoint Zero packet size (both must match)
#define USB_CONTROL_EP_MAX_PACKET_SIZE 64
#define USB_CONTROL_EP_SIZE USB_EP_SIZE_64
//@}

/**
 @name Device Configuration Areas
 @brief Size and location of string descriptors and
 structure used to modify device and configuration
 descriptors.
 */
//@{
// String descriptors - allow a maximum of 256 bytes for this
#define STRING_DESCRIPTOR_LOCATION 0x80
#define STRING_DESCRIPTOR_ALLOCATION 0x100
//@}

/**
 @name DFU_BLOCK_SIZE definition.
 @brief Number of bytes in block, sent in each DFU_DNLOAD request
 from the DFU update program on the host. This is simplified
 in that the meaning of a block is an arbitrary number of
 bytes. This is intentionally a multiple of the maximum
 packet size for the control endpoints.
 */
//@{
#define DFU_BLOCK_SIZE 64

/**
 @name USB_PID_KEYBOARD configuration.
 @brief Run Time Product ID for mouse function.
 */
//@{
#define USB_PID_MOUSE 0x0fea
//@}

/**
 @name DFU_USB_PID_DFUMODE configuration.
 @brief FTDI predefined DFU Mode Product ID.
 */
//@{
#define DFU_USB_PID_DFUMODE 0x0fee
//@}

/**
 @name DFU_USB_INTERFACE configuration..
 @brief Run Time and DFU Mode Interface Numbers.
 */
//@{
#define DFU_USB_INTERFACE_RUNTIME 1
#define DFU_USB_INTERFACE_DFUMODE 0
//@}

/**
 @name WCID_VENDOR_REQUEST_CODE for WCID.
 @brief Unique vendor request code for WCID OS Vendor Extension validation.
*/
//@{
#define WCID_VENDOR_REQUEST_CODE	 0xF1
//@}

/**
 @def MICROSECONDS_TO_TIMER_TICKS
 @brief Macro to calculate the number of timer ticks required to
 create a 1ms counter.
 Scale is clock frequency in MHz (assume 48) / clock-cycles per instruction.
 16-bit timer counts up to 65536, so subtract duration from 65536.
 */
#define MICROSECONDS_TO_TIMER_TICKS(a) (65536 - ((a) * (48 / 12)))

/**
 @name int_in
 @brief Interrupt IN endpoint setup status.
 */
FT51_STATUS int_in = FT51_FAILED;

/**
 @name report_protocol
 @brief Flag to set report protocol (0 - boot, 1 - report)
 */
uint8_t report_protocol = 0;

/**
 @name string_descriptor
 @brief Table of USB String descriptors

 This is placed at a fixed location in the __code section allowing
 up-to 256 bytes of string descriptors to be defined. These can be
 modified or replaced by a utility or binary editor without
 requiring a recompilation of the firmware.
 They are placed at offset 0x100 and reserve 0x100 bytes.
 The data is not stored in __xdata section and is therefore
 regarded as const.
 */
__code __at(STRING_DESCRIPTOR_LOCATION)uint8_t string_descriptor[STRING_DESCRIPTOR_ALLOCATION] =
{
    // String 0 is actually a list of language IDs supported by the real strings.
    0x04, USB_DESCRIPTOR_TYPE_STRING, 0x09,
    0x04, // 0409 = English (US)

    // String 1 (Manufacturer): "FTDI"
    0x0a, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, 'D', 0x00, 'I',
    0x00,

    // String 2 (Product): "FT51A Mouse"
    0x18, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '5', 0x00, '1',
    0x00, 'A', 0x00, ' ', 0x00, 'M', 0x00, 'o', 0x00, 'u', 0x00, 's', 0x00,
    'e', 0x00,

    // String 3 (Serial Number): "FT424242"
    0x12, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '4', 0x00, '2',
    0x00, '4', 0x00, '2', 0x00, '4', 0x00, '3', 0x00,

    // String 4 (DFU Product Name): "FT51A DFU Mode"
    0x1E, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '5', 0x00, '1',
    0x00, 'A', 0x00, ' ', 0x00, 'D', 0x00, 'F', 0x00, 'U', 0x00, ' ', 0x00,
    'M', 0x00, 'o', 0x00, 'd', 0x00, 'e', 0x00,

    // String 5 (Interface Name): "DFU Interface"
    0x1c, USB_DESCRIPTOR_TYPE_STRING, 'D', 0x00, 'F', 0x00, 'U', 0x00, ' ',
    0x00, 'I', 0x00, 'n', 0x00, 't', 0x00, 'e', 0x00, 'r', 0x00, 'f', 0x00,
    'a', 0x00, 'c', 0x00, 'e', 0x00,

    // END OF STRINGS
    0x00
};

/**
 @name wcid_string_descriptor
 @brief USB String descriptor for WCID identification.
 */
__code uint8_t wcid_string_descriptor[USB_MICROSOFT_WCID_STRING_LENGTH] = {
		USB_MICROSOFT_WCID_STRING(WCID_VENDOR_REQUEST_CODE)
};

/**
 @name device_descriptor_mouse
 @brief Device descriptor for Run Time mode.
 */
__code USB_device_descriptor device_descriptor_mouse =
{
    .bLength = 0x12,
    .bDescriptorType = USB_DESCRIPTOR_TYPE_DEVICE,
    .bcdUSB = USB_BCD_VERSION_2_0,          // V2.0
    .bDeviceClass = USB_CLASS_DEVICE,       // Defined in interface
    .bDeviceSubClass = USB_SUBCLASS_DEVICE, // Defined in interface
    .bDeviceProtocol = USB_PROTOCOL_DEVICE, // Defined in interface
    .bMaxPacketSize0 = USB_CONTROL_EP_MAX_PACKET_SIZE,
    .idVendor = USB_VID_FTDI,               // idVendor: 0x0403 (FTDI)
    .idProduct = USB_PID_MOUSE,             // idProduct: 0x0fea
    .bcdDevice = 0x0101,                    // 1.1
    .iManufacturer = 0x01,                  // Manufacturer
    .iProduct = 0x02,                       // Product
    .iSerialNumber = 0x03,                  // Serial Number
    .bNumConfigurations = 0x01
};


/**
 @name config_descriptor_mouse
 @brief Config descriptor for Run Time mode.
 */
//@{
__code struct config_descriptor_mouse
{
    USB_configuration_descriptor  configuration;
    USB_interface_descriptor      interface;
    USB_hid_descriptor            hid;
    USB_endpoint_descriptor       endpoint;
    USB_interface_descriptor      dfu_interface;
    USB_dfu_functional_descriptor dfu_functional;
};


// Fields marked with * are updated with information in device_config[]
struct config_descriptor_mouse config_descriptor_mouse =
{
    .configuration.bLength = 0x09,
    .configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_CONFIGURATION,
    .configuration.wTotalLength = sizeof(struct config_descriptor_mouse),
    .configuration.bNumInterfaces = 0x02,
    .configuration.bConfigurationValue = 0x01,
    .configuration.iConfiguration = 0x00,
    .configuration.bmAttributes =  USB_CONFIG_BMATTRIBUTES_VALUE,
    .configuration.bMaxPower = 0xFA,           // 500mA

    // ---- INTERFACE DESCRIPTOR for Mouse ----
    .interface.bLength = 0x09,
    .interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .interface.bInterfaceNumber = 0,
    .interface.bAlternateSetting = 0x00,
    .interface.bNumEndpoints = 0x01,
    .interface.bInterfaceClass = USB_CLASS_HID,             // HID Class
    .interface.bInterfaceSubClass = USB_SUBCLASS_HID_BOOT_INTERFACE, // Boot Protocol
    .interface.bInterfaceProtocol = USB_PROTOCOL_HID_MOUSE,	// Mouse
    .interface.iInterface = 0x02,	                        // iInterface: "FT51A Mouse"

    // ---- HID DESCRIPTOR for Mouse ----
    .hid.bLength             = 0x09,
    .hid.bDescriptorType     = USB_DESCRIPTOR_TYPE_HID,
    .hid.bcdHID              = USB_BCD_VERSION_HID_1_1,
    .hid.bCountryCode        = USB_HID_LANG_NOT_SUPPORTED,
    .hid.bNumDescriptors     = 0x01,
    .hid.bDescriptorType_0   = USB_DESCRIPTOR_TYPE_REPORT,
    .hid.wDescriptorLength_0 = 0x0034,

    // ---- ENDPOINT DESCRIPTOR for Mouse ----
    .endpoint.bLength = 0x07,
    .endpoint.bDescriptorType = USB_DESCRIPTOR_TYPE_ENDPOINT,
    .endpoint.bEndpointAddress = 0x81,
    .endpoint.bmAttributes = USB_ENDPOINT_DESCRIPTOR_ATTR_INTERRUPT,
    .endpoint.wMaxPacketSize = 0x0008,
    .endpoint.bInterval = 0x0a,

    // ---- INTERFACE DESCRIPTOR for DFU Interface ----
    .dfu_interface.bLength = 0x09,
    .dfu_interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .dfu_interface.bInterfaceNumber = DFU_USB_INTERFACE_RUNTIME,
    .dfu_interface.bAlternateSetting = 0x00,
    .dfu_interface.bNumEndpoints = 0x00,
    .dfu_interface.bInterfaceClass = USB_CLASS_APPLICATION,       // Application Specific Class
    .dfu_interface.bInterfaceSubClass = USB_SUBCLASS_DFU,         // Device Firmware Update
    .dfu_interface.bInterfaceProtocol = USB_PROTOCOL_DFU_RUNTIME, // Runtime Protocol
    .dfu_interface.iInterface = 0x05,                             // "DFU Interface"

    // ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
    .dfu_functional.bLength = 0x09,
    .dfu_functional.bDescriptorType = USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL,
    .dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD | USB_DFU_BMATTRIBUTES_WILLDETACH, // bitCanDnload and bitWillDetach
    .dfu_functional.wDetatchTimeOut = DFU_TIMEOUT, // suggest 8192ms
    .dfu_functional.wTransferSize = DFU_BLOCK_SIZE, // make it 16 bytes
    .dfu_functional.bcdDfuVersion = USB_BCD_VERSION_DFU_1_1, // DFU Version 1.1
};
//@}

/**
 @name wcid_feature_runtime
 @brief WCID Compatible ID for DFU interface in runtime.
 */
//@{
__code USB_WCID_feature_descriptor wcid_feature_runtime =
{
	sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
	1, /* bCount */
	{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
	DFU_USB_INTERFACE_RUNTIME, /* bFirstInterfaceNumber */
	1, /* rsv2 - set to 1 */
	{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
	{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};
//@}

/**
 @name wcid_feature_dfumode
 @brief WCID Compatible ID for DFU interface in DFU mode.
 */
//@{
__code USB_WCID_feature_descriptor wcid_feature_dfumode =
{
	sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
	1, /* bCount */
	{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
	DFU_USB_INTERFACE_DFUMODE, /* bFirstInterfaceNumber */
	1, /* rsv2 - set to 1 */
	{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
	{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};
//@}

/**
 @name wcid_device_guid_dfu
 @brief WCID Compatible DeviceInterfaceGUIDs for DFU interface.
 */
//@{
__code USB_WCID_extended_property_descriptor wcid_device_guid_dfu =
{
	sizeof(struct USB_WCID_extended_property_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_DEVICE_GUID, /* wIndex */
	1, /* wCount */
	136, /* dwSize */
	USB_MICROSOFT_WCID_PROPERTY_DATATYPE_STRING_MULTI, /* dwPropertyDataType */
	42, /* wPropertyNameLength */
	{'D', 0x00, 'e', 0x00, 'v', 0x00, 'i', 0x00, 'c', 0x00, 'e', 0x00,
	 'I', 0x00, 'n', 0x00, 't', 0x00, 'e', 0x00, 'r', 0x00, 'f', 0x00,
	 'a', 0x00, 'c', 0x00, 'e', 0x00, 'G', 0x00, 'U', 0x00, 'I', 0x00,
	 'D', 0x00, 's', 0x00, 0x00, 0x00 }, /* null terminated PropertyName */
	80, /* dwPropertyDataLength */
	/* GUID for FT51A DFU interfaces "{1eaaaa95-1bd6-49c2-b4ae-286583f61227}" */
	{'{', 0x00, '1', 0x00, 'e', 0x00, 'a', 0x00, 'a', 0x00, 'a', 0x00,
	 'a', 0x00, '9', 0x00, '5', 0x00, '-', 0x00, '1', 0x00, 'b', 0x00,
	 'd', 0x00, '6', 0x00, '-', 0x00, '4', 0x00, '9', 0x00, 'c', 0x00,
	 '2', 0x00, '-', 0x00, 'b', 0x00, '4', 0x00, 'a', 0x00, 'e', 0x00,
	 '-', 0x00, '2', 0x00, '8', 0x00, '6', 0x00, '5', 0x00, '8', 0x00,
	 '3', 0x00, 'f', 0x00, '6', 0x00, '1', 0x00, '2', 0x00, '2', 0x00,
	 '7', 0x00, '}', 0x00,
	 0x00, 0x00, 0x00, 0x00}, /* bPropertyData */
};
//@}

/**
 @name hid_report_descriptor_mouse
 @brief HID Report descriptor for mouse

 See Device Class Definition for Human Interface Devices (HID) Version 1.11
 from USB Implementers Forum USB.org

    0x05, 0x01,        Usage Page: Generic Desktop Controls
    0x09, 0x02,        Usage: Mouse (2)
    0xA1, 0x01,        Collection: Application
    0x09, 0x01,          Usage: Pointer (1)
    0xA1, 0x00,          Collection: Physical
    0x05, 0x09,            Usage Page: Button (9)
    0x19, 0x01,            Usage Minimum: Button 1
    0x29, 0x03,            Usage Maximum: Button 3
    0x15, 0x00,            Logical Minimum: 0
    0x25, 0x01,            Logical Maximum: 1
    0x95, 0x03,            Report Count: 3
    0x75, 0x01,            Report Size: 1
    0x81, 0x02,            Input: Data (2)
    0x95, 0x01,            Report Count: 1
    0x75, 0x05,            Report Size: 5
    0x81, 0x01,            Input: Constant (1)
    0x05, 0x01,            Usage Page: Generic Desktop Controls
    0x09, 0x30,            Usage: X
    0x09, 0x31,            Usage: Y
    0x09, 0x38,            Usage: Wheel
    0x15, 0x81,            Logical Minimum: -127
    0x25, 0x7f,            Logical Maximum: 127
    0x75, 0x08,            Report Size: 8
    0x95, 0x03,            Report Count: 3
    0x81, 0x06,            Input: Data (6)
    0xC0,                End collection
    0xC0               End collection
**/
//@{
__code uint8_t hid_report_descriptor_mouse[52] =
{
    0x05, 0x01, 0x09, 0x02, 0xA1, 0x01, 0x09, 0x01, 0xA1, 0x00,
    0x05, 0x09, 0x19, 0x01, 0x29, 0x03, 0x15, 0x00, 0x25, 0x01,
    0x95, 0x03, 0x75, 0x01, 0x81, 0x02, 0x95, 0x01, 0x75, 0x05,
    0x81, 0x01, 0x05, 0x01, 0x09, 0x30, 0x09, 0x31, 0x09, 0x38,
    0x15, 0x81, 0x25, 0x7f, 0x75, 0x08, 0x95, 0x03, 0x81, 0x06,
    0xC0, 0xC0
};
//@}

/**
 @struct hid_report_structure
 @brief HID Report structure to match
 hid_report_descriptor_keyboard descriptor.
 **/
//@{
typedef struct hid_mouse_report_structure_t
{
    unsigned char mouseB1   : 1;
    unsigned char mouseB2   : 1;
    unsigned char mouseB3   : 1;
    unsigned char mouseResv : 5;
    unsigned char mouseX;
    unsigned char mouseY;
    unsigned char mouseWheel;
} hid_mouse_report_structure_t;
//@}

/**
 @name report_buffer
 @brief Buffer used to report back to host the current state of mouse direction/speed.
 */
hid_mouse_report_structure_t report_buffer;

/**
 @name device_descriptor_dfumode
 @brief Device descriptor for DFU Mode.
 */
__code USB_device_descriptor device_descriptor_dfumode =
{
    .bLength = 0x12,
    .bDescriptorType    = USB_DESCRIPTOR_TYPE_DEVICE,
    .bcdUSB             = USB_BCD_VERSION_2_0,            // V2.00
    .bDeviceClass       = USB_CLASS_DEVICE,               // Defined in interface
    .bDeviceSubClass    = USB_SUBCLASS_DEVICE,            // Defined in interface
    .bDeviceProtocol    = USB_PROTOCOL_DEVICE,            // Defined in interface
    .bMaxPacketSize0    = USB_CONTROL_EP_MAX_PACKET_SIZE, // 8
    .idVendor           = USB_VID_FTDI,                   // idVendor: 0x0403 (FTDI)
    .idProduct          = DFU_USB_PID_DFUMODE,            // idProduct: 0x0fee
    .bcdDevice          = 0x0101,                         // 1.1
    .iManufacturer      = 0x01,                           // Manufacturer
    .iProduct           = 0x04,                           // Product
    .iSerialNumber      = 0x03,                           // Serial Number
    .bNumConfigurations = 0x01
};

/**
 @name config_descriptor_dfumode
 @brief Config descriptor for DFU Mode.
 */
//@{
__code struct config_descriptor_dfumode
{
    USB_configuration_descriptor configuration;
    USB_interface_descriptor dfu_interface;
    USB_dfu_functional_descriptor dfu_functional;
};

// Fields marked with * are updated with information in device_config[]
struct config_descriptor_dfumode config_descriptor_dfumode =
{
    .configuration.bLength = 0x09,
    .configuration.bDescriptorType     = USB_DESCRIPTOR_TYPE_CONFIGURATION,
    .configuration.wTotalLength        = sizeof(struct config_descriptor_dfumode),
    .configuration.bNumInterfaces      = 0x01,
    .configuration.bConfigurationValue = 0x01,
    .configuration.iConfiguration      = 0x00,
    .configuration.bmAttributes        = USB_CONFIG_BMATTRIBUTES_VALUE,
    .configuration.bMaxPower           = 0xFA, // 500mA

    // ---- INTERFACE DESCRIPTOR for DFU Interface ----
    .dfu_interface.bLength            = 0x09,
    .dfu_interface.bDescriptorType    = USB_DESCRIPTOR_TYPE_INTERFACE,
    .dfu_interface.bInterfaceNumber   = DFU_USB_INTERFACE_DFUMODE,
    .dfu_interface.bAlternateSetting  = 0x00,
    .dfu_interface.bNumEndpoints      = 0x00,
    .dfu_interface.bInterfaceClass    = USB_CLASS_APPLICATION,    // Application Specific Class
    .dfu_interface.bInterfaceSubClass = USB_SUBCLASS_DFU,         // Device Firmware Update
    .dfu_interface.bInterfaceProtocol = USB_PROTOCOL_DFU_DFUMODE, // Runtime Protocol
    .dfu_interface.iInterface         = 0x05,

    // ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
    .dfu_functional.bLength         = 0x09,
    .dfu_functional.bDescriptorType = USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL,
    .dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD | USB_DFU_BMATTRIBUTES_WILLDETACH, // bitCanDnload and bitWillDetach
    .dfu_functional.wDetatchTimeOut = DFU_TIMEOUT,             // suggest 8192ms
    .dfu_functional.wTransferSize   = DFU_BLOCK_SIZE,          // 16
    .dfu_functional.bcdDfuVersion   = USB_BCD_VERSION_DFU_1_1, // DFU Version 1.1
};
//@}

// Forward declarations.
void usb_detach_attach(void);
void usb_setup(void);

/**
 @brief      Invoked by ISR for Timer0 every millisecond
 @details    Each millisecond Timer0 interrupt handler will call
 this routine.
 When the detach counter is active (> zero) then we will decrement it.
 If we are in the DFUSTATE_APPDETACH state when it reaches zero then
 we will assume that the host has not send the USB reset required to
 put the device into DFU Mode and revert the state machine back to
 DFUSTATE_APPIDLE to wait for another DFU_DETACH request.
 We also make a timer for the keypress simulator.
 @param[in]  flags (not used).
 **/
void ms_timer_interrupt(const uint8_t flags)
{
    (void) flags; // Flags not currently used

    // decrement millisecond timers ...
    ms_timer--;
    mouse_timer--;

    // make note of current button states
    // 0x00 = no keys, 0x01 = SW1 only (left), 0x02 = SW2 only (right), 0x03 = both keys
	mouse_button = ((~P0) >> 1) & 0x03;

    // The DFU detach timer must be called once per millisecond
    dfu_timer();

    // Reload the timer
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));
}

/**
 @brief      Initialise Timer0 to fire every millisecond.
 @details    Associate the ms_timer_interrupt handler function with
 the Timer0 interrupt.
 **/
void ms_timer_initialise(void)
{
    // Register our own handler for interrupts from Timer 0
    interrupts_register(ms_timer_interrupt, interrupts_timer0);

    // Timer0 is controlled by TMOD bits 0 to 3, and TCON bits 4 to 5.
    TMOD &= 0xF0; // Clear Timer0 bits
    TMOD |= 0x01; // Put Timer0 in mode 1 (16 bit)

    // Set the count-up value so that it rolls over to 0 after 1 millisecond.
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));

    TCON &= 0xCF; // Clear Timer0's Overflow and Run flags
    TCON |= 0x10; // Start Timer0 (set its Run flag)
}

/**
 @brief      USB class request handler
 @details    Handle class requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and the DFU state machine and status updated as
 required.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS class_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;
    uint8_t interface = LSB(req->wIndex) & 0x0F;

    // For DFU requests the SETUP packet must be addressed to the
    // the DFU interface on the device.

    // For DFU requests ensure the recipient is an interface...
    if ((req->bmRequestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
    USB_BMREQUESTTYPE_RECIPIENT_INTERFACE)
    {
        // ...and that the interface is the correct Runtime interface
        if (dfu_is_runtime())
        {
            if ((interface == DFU_USB_INTERFACE_RUNTIME))
            {
                // Handle only DFU_DETATCH, DFU_GETSTATE and DFU_GETSTATUS
                // when in Runtime mode. Table 3.2 DFU_DETACH is mandatory
                // in Runtime mode, DFU_GETSTATE and DFU_GETSTATUS are
                // optional.
                switch (req->bRequest)
                {
                case USB_CLASS_REQUEST_DETACH:
                    dfu_class_req_detach(req->wValue);
                    usb_detach_attach();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATUS:
                    dfu_class_req_getstatus();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATE:
                    dfu_class_req_getstate();
                    status = FT51_OK;
                    break;
                }
            }
            else
            {
                // Handle HID class requests
                switch (req->bRequest)
                {
                case USB_HID_REQUEST_CODE_SET_IDLE:
                    USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_GET_REPORT:
					USB_transfer(1, USB_DIR_IN, (char *) &report_buffer, sizeof(report_buffer));
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_SET_PROTOCOL:
                    USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                    report_protocol = req->wValue & 0xff;
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_GET_PROTOCOL:
                    USB_transfer(USB_EP_0, USB_DIR_IN, &report_protocol, 1);
                    status = FT51_OK;
                    break;
                }
            }
        }
        // ...or the correct DFU Mode interface
        else
        {
            if (interface == DFU_USB_INTERFACE_DFUMODE)
            {
                // Handle remaining DFU class requests when in DFU Mode.
                // Table 3.2 DFU_DETACH is not supported in DFU Mode.
                switch (req->bRequest)
                {
                case USB_CLASS_REQUEST_DNLOAD:
                    /* For simplicity only program the maximum block size
                     * starting at an address which is an exact multiple
                     * of this. The block size is given in the configuration
                     * descriptor for the DFU interface.
                     */
                    dfu_class_req_download(req->wValue * DFU_BLOCK_SIZE,
                            req->wLength);
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATUS:
                    dfu_class_req_getstatus();
                    if (dfu_is_wait_reset())
                    {
                    	usb_detach_attach();
                    }
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATE:
                    dfu_class_req_getstate();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_CLRSTATUS:
                    dfu_class_req_clrstatus();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_ABORT:
                    dfu_class_req_abort();
                    status = FT51_OK;
                    break;

                default:
                case USB_CLASS_REQUEST_UPLOAD:
                    // Unknown or unsupported request.
                    break;
                }
            }
        }
    }

    return status;
}

/**
 @brief      USB standard request handler for GET_DESCRIPTOR
 @details    Handle standard GET_DESCRIPTOR requests from the host
 application.
 The hValue is parsed and the appropriate device,
 configuration or string descriptor is selected.
 For device and configuration descriptors the DFU
 state machine determines which descriptor to use.
 For string descriptors the lValue selects which
 string from the string_descriptors table to use.
 Strings are obtained from the __code section and
 typecast to a be compatible with __xdata storage.
 The string table is automatically traversed to find
 the correct string and the length is taken from the
 string descriptor.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bmRequestType is not
 supported.
 **/
FT51_STATUS standard_req_get_descriptor(USB_device_request *req)
{
    uint8_t *src = NULL;
    uint16_t length = req->wLength;
    uint8_t hValue = req->wValue >> 8;
    uint8_t lValue = req->wValue & 0x00ff;
    uint8_t i;
    uint8_t slen;

    switch (hValue)
    {
    case USB_DESCRIPTOR_TYPE_DEVICE:
        if (dfu_is_runtime())
        {
            src = (char *) &device_descriptor_mouse;
        }
        else
        {
            src = (char *) &device_descriptor_dfumode;
        }
        if (length > sizeof(USB_device_descriptor)) // too many bytes requested
            length = sizeof(USB_device_descriptor); // Entire structure.
        break;

    case USB_DESCRIPTOR_TYPE_CONFIGURATION:
        if (dfu_is_runtime())
        {
            src = (char *) &config_descriptor_mouse;
            if (length > sizeof(config_descriptor_mouse)) // too many bytes requested
                length = sizeof(config_descriptor_mouse); // Entire structure.
        }
        else
        {
            src = (char *) &config_descriptor_dfumode;
            if (length > sizeof(config_descriptor_dfumode)) // too many bytes requested
                length = sizeof(config_descriptor_dfumode); // Entire structure.
        }
        break;

    case USB_DESCRIPTOR_TYPE_STRING:

		// Special case: WCID descriptor
		if (lValue == USB_MICROSOFT_WCID_STRING_DESCRIPTOR)
		{
			src = (uint8_t *)wcid_string_descriptor;
			length = sizeof(wcid_string_descriptor);
			break;
		}

		// Find the nth string in the string descriptor table
        i = 0;
        while ((slen = string_descriptor[i]) > 0)
        {
            // Point to start of string descriptor in __code segment
            // Typecast prevents warning for losing const when usb_transfer
            // is called
            src = (uint8_t *) &string_descriptor[i];
            if (lValue == 0)
            {
                break;
            }
            i += slen;
            lValue--;
        }
        if (lValue > 0)
        {
            // String not found
            return FT51_FAILED;
        }
        // Update the length returned only if it is less than the requested
        // size
        if (length > slen)
        {
            length = slen;
        }

        break;

    case USB_DESCRIPTOR_TYPE_REPORT:
        src = (char *) &hid_report_descriptor_mouse;
        if (length > sizeof(hid_report_descriptor_mouse))
        {
            length = sizeof(hid_report_descriptor_mouse); // Entire structure.
        }

        break;

    default:
        return FT51_FAILED;
    }

    USB_transfer(USB_EP_0, USB_DIR_IN, src, length);

    return FT51_OK;
}

/**
 @brief      USB standard request handler for GET_STATUS
 @details    Handle standard GET_STATUS requests from the host
             application.
             The bmRequestType value is parsed and the appropriate
             action or function is performed.
 @param[in]	 req - USB_device_request structure containing the
             SETUP portion of the request from the host.
 @return	 status - FT51_OK if successful or FT51_FAILED
             if there is an error or the bmRequestType is not
             supported.
 **/
FT51_STATUS standard_req_get_status(USB_device_request *req)
{
    USB_STATE state = USB_get_state();
    USB_ENDPOINT_NUMBER ep_number = LSB(req->wIndex) & 0x0F;
    USB_ENDPOINT_DIR ep_dir = (LSB(req->wIndex) >> 7);
    uint8_t buf[2];

    buf[0] = buf[1] = 0;

    // Get Status for endpoints only
    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT))
    {
        if (((state < CONFIGURED) && (ep_number == 0))
                || (state >= CONFIGURED))
        {
            if (USB_get_ep_stalled(ep_number, ep_dir))
            {
                buf[0] = USB_GET_STATUS_ENDPOINT_HALT;
            }
            USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

            return FT51_OK;
        }
    }
    else if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_DEVICE))
    {
        // This must match the configuration descriptor's bmAttributes
        buf[0] = USB_CONFIG_BMATTRIBUTES_VALUE;
        USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

        return FT51_OK;
    }
    return FT51_FAILED;
}

/**
 @brief      USB standard request handler for GET_FEATURE and SET_FEATURE
 @details    Handle standard GET_FEATURE and SET_FEATURE requests
             from the host application.
             The bmRequestType value is parsed and the appropriate
             action or function is performed.
 @param[in]  req - USB_device_request structure containing the
             SETUP portion of the request from the host.
 @return        status - FT51_OK if successful or FT51_FAILED
             if there is an error or the bmRequestType is not
             supported.
 **/
FT51_STATUS standard_req_set_clear_feature(USB_device_request *req)
{
    USB_ENDPOINT_NUMBER ep_number = LSB(req->wIndex) & 0x0F;
    USB_ENDPOINT_DIR ep_dir = (LSB(req->wIndex) >> 7);
    USB_STATE state = USB_get_state();

    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_HOST_TO_DEV |
            USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT))
    {
        // Only support the endpoint halt feature in this device
        if (req->wValue == USB_FEATURE_ENDPOINT_HALT)
        {
            // Only allow this if the device is configured or the endpoint
            // is zero if the device is not configured.
            if (((state < CONFIGURED) && (ep_number == 0))
                    || (state >= CONFIGURED))
            {
                // Perform a stall or a clear
                if (req->bRequest == USB_REQUEST_CODE_CLEAR_FEATURE)
                {
                    // Feature selector CLEAR
                    USB_clear_endpoint(ep_number, ep_dir);
                }
                else
                {
                    // Feature selector SET
                    USB_stall_endpoint(ep_number, ep_dir);
                }
                USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                return FT51_OK;
            }
        }
    }
    return FT51_FAILED;
}

/**
 @brief      USB standard request handler
 @details    Handle standard requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS standard_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;

    switch (req->bRequest)
    {
    case USB_REQUEST_CODE_GET_STATUS:
        status = standard_req_get_status(req);
        break;

    case USB_REQUEST_CODE_CLEAR_FEATURE:
    case USB_REQUEST_CODE_SET_FEATURE:
        status = standard_req_set_clear_feature(req);
        break;

    case USB_REQUEST_CODE_GET_DESCRIPTOR:
        status = standard_req_get_descriptor(req);
        break;

    case USB_REQUEST_CODE_SET_CONFIGURATION:
        status = USB_req_set_configuration(req);
        break;

    case USB_REQUEST_CODE_GET_CONFIGURATION:
        status = USB_req_get_configuration();
        break;

    case USB_REQUEST_CODE_SET_ADDRESS:
        status =  USB_req_set_address(req);
        break;

    default:
    case USB_REQUEST_CODE_GET_INTERFACE:
    case USB_REQUEST_CODE_SET_INTERFACE:
        // Unknown or unsupported request.
        status = FT51_FAILED;
        break;
    }

    return status;
}

/**
 @brief      USB vendor request handler
 @details    Handle vendor requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 There are no vendor requests requiring handling in
 this firmware.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS vendor_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;
	uint16_t length = req->wLength;

	// For Microsoft WCID only.
	// Request for Compatible ID Feature Descriptors.
	// Check the request code and wIndex.
	if (req->bRequest == WCID_VENDOR_REQUEST_CODE)
	{
		if (req->bmRequestType & USB_BMREQUESTTYPE_DIR_DEV_TO_HOST)
		{
			switch (req->wIndex)
			{
			case USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID:
				if (length > sizeof(wcid_feature_runtime)) // too many bytes requested
					length = sizeof(wcid_feature_runtime); // Entire structure.
				// Return a compatible ID feature descriptor.
				if (dfu_is_runtime())
				{
					USB_transfer(USB_EP_0, USB_DIR_IN,
						(uint8_t *) &wcid_feature_runtime, length);
					status = FT51_OK;
				}
				else
				{
					USB_transfer(USB_EP_0, USB_DIR_IN,
						(uint8_t *) &wcid_feature_dfumode, length);
					status = FT51_OK;
				}
				break;

			case USB_MICROSOFT_WCID_FEATURE_WINDEX_DEVICE_GUID:
				if (length > sizeof(wcid_device_guid_dfu)) // too many bytes requested
					length = sizeof(wcid_device_guid_dfu); // Entire structure.
				// Return a compatible ID feature descriptor.
				if (((dfu_is_runtime()) && (req->wValue == DFU_USB_INTERFACE_RUNTIME))
					|| ((!dfu_is_runtime()) && (req->wValue == DFU_USB_INTERFACE_DFUMODE)))
				{
					USB_transfer(USB_EP_0, USB_DIR_IN,
							(uint8_t *) &wcid_device_guid_dfu, length);
					status = FT51_OK;
				}
				break;
			}
		}
	}
    return status;
}

/**
 @brief      USB suspend callback
 @details    Called when the USB bus enters the suspend state.
 @param[in]	status - unused.
 @return		N/A
 **/
void suspend_cb(uint8_t status)
{
    (void) status;
    return;
}

/**
 @brief      USB resume callback
 @details    Called when the USB bus enters the resume state
 prior to restating after a suspend.
 @param[in]  status - unused.
 @return     N/S
 **/
void resume_cb(uint8_t status)
{
    (void) status;
    return;
}

/**
 @brief      USB reset callback
 @details    Called when the USB bus enters the reset state.
 @param[in]	status - unused.
 @return		N/A
 **/
void reset_cb(uint8_t status)
{
    (void) status;

    USB_free_endpoint(1, USB_DIR_IN);

    USB_set_state(DEFAULT);

    if (dfu_reset())
    {
        // Run downloaded application - reload MTP to shadow RAM
        // Execution effectively stops now and returns to the main()
        // function of the new program in MTP.
        device_revert();
    }

    /* Make application specific BULK, INTERRUPT or
     * ISOCHRONOUS endpoints.
     */
    int_in = USB_create_endpoint(USB_EP_1, USB_EP_INT,
            USB_DIR_IN, USB_EP_SIZE_8);

    return;
}

/**
 @brief      USB detach attach sequence
 @details    Called when the USB device wishes to detach from the
 	 	 	 bus then reattach.
 @return		N/A
 **/
void usb_detach_attach(void)
{
	ms_timer = 10;
	while (ms_timer > 0);
	// Disable USB device to force an reset
	USB_finalise();
	ms_timer = 10;
	while (ms_timer > 0);
	// Start the USB device.
	usb_setup();
}

/**
 @brief      Initialise emulation of a USB device
 @details    Sets up callback functions and data structures
 required to implement control endpoints on device.
 @param[in]	ep_size - Endpoint size of control endpoints.
 @return     N/A
 **/
void usb_setup(void)
{
    USB_ctx usb_ctx;

    memset(&usb_ctx, 0, sizeof(usb_ctx));

    usb_ctx.standard_req = standard_req_cb;
    usb_ctx.class_req = class_req_cb;
    usb_ctx.vendor_req = vendor_req_cb;
    usb_ctx.suspend = suspend_cb;
    usb_ctx.resume = resume_cb;
    usb_ctx.reset = reset_cb;
    usb_ctx.lpm = NULL;

    usb_ctx.hub_en.mask.hub_enable = USB_HUB_ENABLED;
    usb_ctx.hub_en.mask.hub_LocalPwrSrc = USB_SELF_POWERED;
    usb_ctx.hub_en.mask.hub_RmtWkUpEn = 1;

    // Initialise the USB device with a control endpoint size
    // of 8 bytes. This must match the size for bMaxPacketSize0
    // defined in the device descriptor.
    usb_ctx.ep_size = USB_CONTROL_EP_SIZE;

    // Perform USB Device initialisation.
    USB_initialise(&usb_ctx);

    return;
}

/**
 @brief      Initialise the IOMUX for the input and output pins.
 @details    Set the input and output characteristics for the pins used
 	         on the FT51A EVM module.
 	         I2C Master SCL is on pin DIO_13
 	         I2C Master SDA is on pin DIO_14.
 	         LCD reset line is on DIO_8 and mapped to P0.0
 	         SW1 is on DIO_0 and mapped to P0.1
 	         SW2 is on DIO_7 and mapped to P0.2
 **/
void iomux_setup()
{
	// LCD control pin
    IOMUX_OUTPUT(DIO_8,  IOMUX_OUT_PORT0_0);

    // I2C Master
    IOMUX_OUTPUT(DIO_13, IOMUX_OUT_I2C_MASTER_SCL);
    IOMUX_OUTPUT(DIO_14, IOMUX_OUT_I2C_MASTER_SDA);
    IOMUX_INPUT (DIO_14, IOMUX_IN_I2C_MASTER_SDA);

    IOMUX_set_pad_config(DIO_8,IOMUX_PAD_DRIVE_CURRENT_16MA,
                            IOMUX_PAD_TRIGGER_SCHMITT,
                            IOMUX_PAD_SLEW_RATE_FAST,
                            IOMUX_PAD_PULL_NONE);
    IOMUX_set_pad_config(DIO_13,IOMUX_PAD_DRIVE_CURRENT_8MA,
                            IOMUX_PAD_TRIGGER_SCHMITT,
                            IOMUX_PAD_SLEW_RATE_FAST,
                            IOMUX_PAD_PULL_NONE);
    IOMUX_set_pad_config(DIO_14,IOMUX_PAD_DRIVE_CURRENT_4MA,
                            IOMUX_PAD_TRIGGER_SCHMITT,
                            IOMUX_PAD_SLEW_RATE_FAST,
                            IOMUX_PAD_PULL_NONE);

    // Setup switches
    // 0x00 = no keys, 0x01 = SW1 only (left), 0x02 = SW2 only (right), 0x03 = both keys
    IOMUX_INPUT(DIO_0, IOMUX_IN_PORT0_1); // SW1
    IOMUX_INPUT(DIO_7, IOMUX_IN_PORT0_2); // SW2
}

void waitforawhile(int16_t wait_time)
{
    ms_timer = wait_time;
    while (ms_timer);
}

/**
 @brief      Interrupt handler for ADC conversion completion.
 @details    Update the value of the force applied to the force sensor
 	 	 	 when an interrupt is received indicating that the
 	 	 	 conversion process has completed.
 **/
void adc_interrupt(const uint8_t flags)
{
    uint8_t interrupt = 0;
    uint8_t sample_l = 0x00;
    uint8_t sample_h = 0x00;

    (void)flags;

    DISABLE_INTERRUPTS;

    // Determine the interrupt source.
    IO_REG_GENERAL_READ(AIO_INTERRUPTS_0_7, interrupt);
    // FSR Output
    if (interrupt & MASK_AIO_0_INT)
    {
        IO_REG_GENERAL_READ(AIO_0_ADC_DATA_L_1, sample_l);
        IO_REG_GENERAL_READ(AIO_0_ADC_DATA_U_1, sample_h);

        mouse_force = ((sample_h << 8) | sample_l);

        // Clear ADC interrupt register bit
        IO_REG_GENERAL_WRITE(AIO_INTERRUPTS_0_7, MASK_AIO_0_INT);
    }

    ENABLE_INTERRUPTS;
}

/**
 @brief      main function.
 @details    Modify any configuration parameters which are updatable
             and call initialisation routines.
             Finally process USB requests as they arrive (for ever).
 **/
void main(void)
{
	uint8_t mouse_motion = 0; // direction in which switch presses will move mouse pointer
	uint8_t last_pressed = 0; // record of last button press
	uint8_t next_pressed = 0; // record of next button press
	uint8_t need_release = 0; // when both button are pressed to change direction they must be released before continuing
	uint8_t speed = 0;        // current speed of mouse movement (increases if same button is held down)
	uint8_t force = 0;        // force read by sensor (if above preset limit then mouse button pressed)
	uint8_t lcdtext[21];      // temp buffer to store text for display on LCD

    // Initialise FT51A
    device_initialise();

    // Initialise IO
    iomux_setup();

    // Enable Force ADC ...
    // Enable AIO Cell Controller
    IO_REG_GENERAL_WRITE(AIO_CONTROL, MASK_AIO_DEV_EN);
    // Select mode. Setup AIO_0 to ADC
    IO_REG_GENERAL_WRITE(AIO_MODE_0, MASK_AIO_0_MODE_1);
    // Enable AIO_0 interrupt
    IO_REG_GENERAL_WRITE(AIO_INTERRUPT_ENABLES_0_7, MASK_AIO_0_IEN);
    // Setup AIO_2 to disable pull up/down and drive strength to zero
    IO_REG_GENERAL_WRITE(DIGITAL_CONTROL_AIO_0, 0x00);

    // Register for ADC interrupts
    interrupts_register(adc_interrupt, interrupts_IO_cell);

    // Initialise USB function
    usb_setup();

    memset(&report_buffer, 0, sizeof(report_buffer));

    // Initialise the timer to provide a millisecond countdown timer
    ms_timer_initialise();

    I2C_master_initialise();

    // Reset the LCD display ...
    P0 = 0xf8;        // Set reset line of LCD high
    waitforawhile(5); // Wait for 5ms
    P0 = 0xf9;        // Clear reset line of LCD

    // Wait 40ms after reset
    waitforawhile(40);

    // Initialise and clear the display
    lcd_initialise();
    lcd_clear();

    // Wait for 2ms for clear to complete
    waitforawhile(2);

    // Endless loop.
    while (1)
    {
        // Check control endpoints and handle SETUPs
        USB_process();

        // mouse X, Y and speed emulation code
        if (USB_get_state() == CONFIGURED)
        {
        	// send mouse report periodically (every MOUSE_REPORT_PERIOD milliseconds)
        	if (mouse_timer < 0)
        	{
        		if (!USB_ep_buffer_full(1, USB_DIR_IN))
        		{
    		        // Initiate ADC sample AIO_0 (to read force sensor)
    		        IO_REG_GENERAL_WRITE(AIO_SAMPLE_0,  MASK_AIO_0_SAMPLE);

    		        // reset mouse report to default values
        			report_buffer.mouseB1    = 0;
            		report_buffer.mouseB2    = 0;
            		report_buffer.mouseB3    = 0;
            		report_buffer.mouseWheel = 0;
            		report_buffer.mouseX     = 0;
            		report_buffer.mouseY     = 0;

            	    DISABLE_INTERRUPTS;

            	    // record last and next button presses (used to increase speed if same button is held down)
            	    last_pressed = next_pressed;
            	    next_pressed = mouse_button;

            	    // max force reading seems to be nearly 900 (scale to 0-8 range)
            	    force = mouse_force / 100;

            	    // delay before sending next mouse report
					mouse_timer = MOUSE_REPORT_PERIOD;

            	    ENABLE_INTERRUPTS;

        			if (need_release)
        			{
        				need_release = next_pressed;
        			}
            	    else
            	    {
            	    	// consider mouse button pressed if force level is high enough
                		if (force > 4)
                		{
                			report_buffer.mouseB1 = 1;
                		}

                		// increase speed of movement if button is held down
						if (last_pressed == next_pressed)
						{
							if (speed < MOUSE_SPEED_LIMIT)
							{
								speed++;
							}
						}
						else
						{
							speed = 0;
						}


            	    	// Clear display and wait for 2ms for clear to complete
                	    lcd_clear();
                	    waitforawhile(2);

                		switch (next_pressed)
                		{
                		case 0x00:
                			if (mouse_motion)
                			{
                        	    lcd_display("[U]   <-Move->   [D]", 20);
                			}
                			else
                			{
                        	    lcd_display("[L]   <-Move->   [R]", 20);
                			}
                			speed = 0;
                			break;

                		case 0x01:
							if (mouse_motion == 0x00)
							{
								report_buffer.mouseX = 0xFF - (speed / MOUSE_SPEED_SCALE); // move left
								lcd_display("[L]<--- Move     [R]", 20);
							}
							else
							{
								report_buffer.mouseY = 0xFF - (speed / MOUSE_SPEED_SCALE); // move up
								lcd_display("[U]<--- Move     [D]", 20);
							}
                    		break;

    					case 0x02:
							if (mouse_motion == 0x00)
							{
								report_buffer.mouseX = 0x01 + (speed / MOUSE_SPEED_SCALE); // move right
								lcd_display("[L]     Move --->[R]", 20);
							}
							else
							{
								report_buffer.mouseY = 0x01 + (speed / MOUSE_SPEED_SCALE); // move down
								lcd_display("[U]     Move --->[D]", 20);
							}
							break;

    					case 0x03:
							mouse_motion = !mouse_motion; // switch direction between left/right and up/down
							need_release = 1;             // keys need to be released after changing direction
							lcd_display(" Release both keys! ", 20);
    						break;
                		}

                		lcd_position(1, 0);
            			sprintf(lcdtext, "Force[%1d]   Speed[%02d]", force, speed / MOUSE_SPEED_SCALE);
            			lcd_display(lcdtext, 20);
            	    }

        			// refresh mouse pointer position
					USB_transfer(1, USB_DIR_IN, (char *) &report_buffer, sizeof(report_buffer));
        		}
        	}
        }
    }
}
